//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#pragma once
#include "types.h"
#include "opcodes.h"
#include "updownclient.h"
#include "preferences.h"
#include "loggable.h"

class CEdt; //<<--enkeyDev(th1) -EDT-

class CUploadQueue: public CLoggable
{
	friend class CEdt; //<<--enkeyDev(th1) -EDT-
public:
	CUploadQueue(CPreferences* in_prefs);
	~CUploadQueue();
	void	ProcessSending();
	void	AddClientToQueue(CUpDownClient* client,bool bIgnoreTimelimit = false);
	bool	RemoveFromUploadQueue(CUpDownClient* client,bool updatewindow = true);
	bool	RemoveFromWaitingQueue(CUpDownClient* client,bool updatewindow = true);
	bool	IsOnUploadQueue(CUpDownClient* client)	{CSingleLock waitinglistLock(&waitinglistMutex); waitinglistLock.Lock(); bool is=(waitinglist.Find(client) != 0); waitinglistLock.Unlock(); return is;} //Pawcio: BC
	bool	IsDownloading(CUpDownClient* client)	const {return (client->GetUploadState() == US_UPLOADING || client->GetUploadState() == US_CONNECTING);/*(uploadinglist.Find(client) != 0 || reserveduploadinglist.Find(client) !=0);*/} // Pawcio: BC
	uint32	GetDatarate()							{return datarate;}
	bool	CheckForTimeOver(CUpDownClient* client, bool fullChunk=false); // Pawcio: BC
	int		GetWaitingUserCount()					{CSingleLock waitinglistLock(&waitinglistMutex); waitinglistLock.Lock(); int count=waitinglist.GetCount(); waitinglistLock.Unlock(); return count;} //Pawcio: BC
	int		GetUploadQueueLength()					{return uploadinglist.GetCount();}
	
	//Pawcio: BC
	void	WaitForLoadingParts();
	void	AddToBlockedQueue(CEMSocket* const & sock);
	void	AddToControlQueue(CEMSocket* const & sock);
	void	DecreaseToSend(const int& bytes);
	//void    RemoveFromBlockedSockets(CEMSocket* const & sock);
	//void	RemoveFromEMSockQueue(CEMSocket* const & sock);
	uint16	SlotsInRange(bool lowID = false);
	uint16  MaxSlots(bool lowID = false);
	bool	AcceptToReserved(bool lowID = false);
	bool	RemoveFromReservedQueue(CUpDownClient* client,bool updatewindow = true);
	uint16	GetReservedQueueCount() { return reserveduploadinglist.GetCount(); }
	uint32	GetHighestScoreOnReserved() {return highestScoreOnReserved;}
	bool	IsSocketOnBlocked(CEMSocket* const & sock){ return blockedSockets.Find(sock);}
	bool	IsSocketOnControl(CEMSocket* const & sock){	return EMSockControlQueue.Find(sock);}
	CUpDownClient* GetPriorityClient() {return priorityClient;}
	void	ResetBC() {feedingPos = NULL; reservedFeedingPos = NULL; /*priorityClient = NULL;*/ feeding = NULL;}
	//Pawcio: BC

	POSITION GetFirstFromUploadList()				{return uploadinglist.GetHeadPosition();}
	CUpDownClient* GetNextFromUploadList(POSITION &curpos)	{return uploadinglist.GetNext(curpos);}
	CUpDownClient* GetQueueClientAt(POSITION &curpos)	{return uploadinglist.GetAt(curpos);}

	POSITION GetFirstFromWaitingList()				{CSingleLock waitinglistLock(&waitinglistMutex); waitinglistLock.Lock(); POSITION pos=waitinglist.GetHeadPosition(); waitinglistLock.Unlock(); return pos;} //Pawcio: BC
	CUpDownClient* GetNextFromWaitingList(POSITION &curpos)	{CSingleLock waitinglistLock(&waitinglistMutex); waitinglistLock.Lock(); CUpDownClient* client=waitinglist.GetNext(curpos); waitinglistLock.Unlock(); return client;} //Pawcio: BC
	CUpDownClient* GetWaitClientAt(POSITION &curpos)	{CSingleLock waitinglistLock(&waitinglistMutex); waitinglistLock.Lock(); CUpDownClient* client=waitinglist.GetAt(curpos); waitinglistLock.Unlock(); return client;} //Pawcio: BC

	CUpDownClient*	GetWaitingClientByIP_UDP(uint32 dwIP, uint16 nUDPPort);
#ifdef _DEBUG
	CUpDownClient*	GetWaitingClientByIP(uint32 dwIP);
#endif
	CUpDownClient*	GetNextClient(CUpDownClient* update);

	
	void	DeleteAll();
	uint16	GetWaitingPosition(CUpDownClient* client);
	
	//void	SetBanCount(uint32 in_count)			{bannedcount = in_count;}
	//uint32	GetBanCount()							{return bannedcount;}
	//void	UpdateBanCount();
	
	uint32	GetSuccessfullUpCount()					{return successfullupcount;}
	uint32	GetFailedUpCount()						{return failedupcount;}
	uint32	GetAverageUpTime();
	void	FindSourcesForFileById(CTypedPtrList<CPtrList, CUpDownClient*>* srclist, const uchar* filehash);
	void	AddUpDataOverheadSourceExchange(uint32 data)	{ m_nUpDataRateMSOverhead += data;
															  m_nUpDataOverheadSourceExchange += data;
															  m_nUpDataOverheadSourceExchangePackets++;}
	void	AddUpDataOverheadFileRequest(uint32 data)		{ m_nUpDataRateMSOverhead += data;
															  m_nUpDataOverheadFileRequest += data;
															  m_nUpDataOverheadFileRequestPackets++;}
	void	AddUpDataOverheadServer(uint32 data)			{ m_nUpDataRateMSOverhead += data;
															  m_nUpDataOverheadServer += data;
															  m_nUpDataOverheadServerPackets++;}
	void	AddUpDataOverheadOther(uint32 data)				{ m_nUpDataRateMSOverhead += data;
															  m_nUpDataOverheadOther += data;
															  m_nUpDataOverheadOtherPackets++;}
	uint32	GetUpDatarateOverhead()						{return m_nUpDatarateOverhead;}
	uint64	GetUpDataOverheadSourceExchange()			{return m_nUpDataOverheadSourceExchange;}
	uint64	GetUpDataOverheadFileRequest()				{return m_nUpDataOverheadFileRequest;}
	uint64	GetUpDataOverheadServer()					{return m_nUpDataOverheadServer;}
	uint64	GetUpDataOverheadOther()					{return m_nUpDataOverheadOther;}
	uint64	GetUpDataOverheadSourceExchangePackets()	{return m_nUpDataOverheadSourceExchangePackets;}
	uint64	GetUpDataOverheadFileRequestPackets()		{return m_nUpDataOverheadFileRequestPackets;}
	uint64	GetUpDataOverheadServerPackets()			{return m_nUpDataOverheadServerPackets;}
	uint64	GetUpDataOverheadOtherPackets()				{return m_nUpDataOverheadOtherPackets;}
	void	CompUpDatarateOverhead();
	void	UpdateMaxClientScore();

protected:
	void	RemoveFromWaitingQueue(POSITION pos, bool updatewindow);
//	POSITION	GetWaitingClient(CUpDownClient* client);
//	POSITION	GetWaitingClientByID(CUpDownClient* client);
//	POSITION	GetDownloadingClient(CUpDownClient* client);
	bool		AcceptNewClient();
	void		AddUpNextClient(CUpDownClient* directadd = 0, bool toReserve=false); //Pawcio: BC
	
	//Pawcio: BC

private:
	uint32	GetMaxClientScore()						{return m_imaxscore;}

	// By BadWolf - Accurate Speed Measurement
	CList<TransferredData,TransferredData> avarage_dr_list;
	DWORD	m_delay;
	DWORD	m_delaytmp;
	uint32	uLastAcceptNewClient;
	uint32	sumavgdata;
	sint32	sendperclient;

	CTypedPtrList<CPtrList, CUpDownClient*> waitinglist;
	CTypedPtrList<CPtrList, CUpDownClient*> uploadinglist;
	uint32	datarate;   //datarate 
	uint32	dataratems;
	uint32	datarateave; //datarage average (since progstart) *unused*
	uint32	estadatarate; // esta. max datarate	

	// Pawcio: BC
	CMutex waitinglistMutex;
	CTypedPtrList<CPtrList, CUpDownClient*> withoutpart;
	CMutex withoutpartMutex;
	CTypedPtrList<CPtrList, CUpDownClient*> reserveduploadinglist;
	CTypedPtrList<CPtrList, CUpDownClient*> prioritylist;
	CTypedPtrList<CPtrList, CEMSocket*> blockedSockets;
	CTypedPtrList<CPtrList, CEMSocket*> EMSockControlQueue;
	CList<TransferredData,TransferredData> sentToNormalList;
	CList<TransferredData,TransferredData> sentToReservedList;
	uint32 DataSentToNormal;
	uint32 DataSentToReserved;
	uint32 MaxUpload;
	uint16 SentToNormal;
	uint16 SentToReserved;
	uint16 clientsrdy;
	uint16 reservedclientsrdy;
	CUpDownClient* feeding;
	CUpDownClient* priorityClient;
	POSITION feedingPos;
	POSITION reservedFeedingPos;
	static UINT LoadingParts(LPVOID pParam);
	uint32  m_lastProcessTime;
	uint32	m_nLastAddToReserved;
	int     toSend;
	uint32	highestScoreOnReserved;
	uint16 endingTransfer;
	bool	bandwidthUsed;
	CEvent*	endLoadingParts;
	// Pawcio: BC

	CPreferences* app_prefs;
	//Pawcio: BC
	uint32	successfullupcount;
	uint32	failedupcount;
	uint32	totaluploadtime;
	uint32	m_nLastStartUpload;
	uint32	m_nUpDatarateOverhead;
	uint32	m_nUpDataRateMSOverhead;
	uint64	m_nUpDataOverheadSourceExchange;
	uint64	m_nUpDataOverheadFileRequest;
	uint64	m_nUpDataOverheadServer;
	uint64	m_nUpDataOverheadOther;
	uint64	m_nUpDataOverheadSourceExchangePackets;
	uint64	m_nUpDataOverheadFileRequestPackets;
	uint64	m_nUpDataOverheadServerPackets;
	uint64	m_nUpDataOverheadOtherPackets;
	bool	lastupslotHighID; // VQB lowID alternation

	// By BadWolf - Accurate Speed Measurement	
	CList<TransferredData,TransferredData>	m_AvarageUDRO_list;
	uint32	sumavgUDRO;
	// END By BadWolf - Accurate Speed Measurement	
	uint32	m_imaxscore;
};
